#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#ifdef ADWIN_MODVERSIONS
#include <linux/modversions.h>
#endif

#include <linux/kernel.h>	/* we want to live in kernel */
#include <linux/module.h>	/* we are a module */
#include <linux/types.h>	/* some independent types */

#include <asm/uaccess.h>	/* access user space */
#include <asm/io.h>		/* inb(), outb() */

#include <linux/adwin.h>

#ifdef ADWIN_DEBUG
  extern int debug;
  #define DEBUG(n, args...) if (debug > (n)) printk(args)
#else
  #define DEBUG(n, args...)
#endif

#include "adwin_link.h"

/*****************************************************************************/
/* Reset the C011 and DSP connection
 * To reset, toggle the bits in the high regions of a base addr
 * (e.g. base := 0x150, toggle 0x160)
 * timeout steps are in jiffies (about 10ms)
 */
int reset_link(wait_queue_head_t *q, int base_addr) {
	DEBUG(1, "ADwin debug: reset_link(0x%X)\n", base_addr);

	outb(0, base_addr + 0x11);
	outb(0, base_addr + 0x10);
	sleep_on_timeout(q, 1);

	outb(1, base_addr + 0x10);
	sleep_on_timeout(q, 2);

	outb(0, base_addr + 0x10);
	sleep_on_timeout(q, 1);

	return 0;
}

/*****************************************************************************/
/* timout signal handler
 * After installing the timer in the to/from-link functions, this function 
 * is called if the timer expires. The flag is set, so the to/from functions
 * will stop their polling and tell the user (ETIMEDOUT).
 */
void adwin_timeout(unsigned long __data) {
	*((long *)__data) = 1;
}

/*****************************************************************************/
/* read from link (transputer)
 * This is still endianess-dependent. It will work on litte-endian systems.
 * I don't change it now, beacuse transputer isn't officially supported any 
 * more. 
 * If this shall work at be-systems, change it according to from_link_dsp.
 */
static int from_link(wait_queue_head_t *q, int base_addr,
						char *usr_buf, size_t cnt) {
	size_t i;
	__u8 *ptr = (__u8 *) usr_buf;
	struct timer_list timer;
	volatile long timeout_flag = 0;
#ifdef ADWIN_DEBUG
	char c;
#endif

	if (! access_ok(VERIFY_WRITE, usr_buf, cnt)) {
		DEBUG(0, "ADwin debug: Memory access error\n");
		return -EFAULT;
	}

	DEBUG(2, "ADwin debug: ");

        while(cnt > 0) {
		/* Start polling. 
		 * In the most cases within this 100 peeks we are allowed
		 * to read.
		 */
		for (i = 0; (i < 100) && READ_STATUS(base_addr); i++);

		if (i < 100) {
			/* Read and go on ... */
#ifdef ADWIN_DEBUG
			c = inb(base_addr);
			DEBUG(2, "%02X ", c);
			__put_user(c, ptr++);
#else
			__put_user(inb(base_addr), ptr++);
#endif
			cnt--;
			continue;
		}

		/* Seems to take while.
		 * Install a timer and go on peeking. */
		init_timer(&timer);
		/* timer.expires  = jiffies + 25; */
		/* read-timeout fr booten der Transputer mit Speicher >= 4MB hochgesetzt! */
		timer.expires  = jiffies + 50;
		timer.data     = (unsigned long) &timeout_flag;
		timer.function = adwin_timeout;

		add_timer(&timer);
		while(! timeout_flag) {
			if (READ_STATUS(base_addr) != -EACCES) {
#ifdef ADWIN_DEBUG
				c = inb(base_addr);
				DEBUG(2, "%02X ", c);
				__put_user(c, ptr++);
#else
				__put_user(inb(base_addr), ptr++);
#endif
				cnt--;
				break;
			} else
				schedule();
		}
		del_timer_sync(&timer);

		if (timeout_flag) {
			DEBUG(1, "\nADwin debug: timed out\n");
			return -ETIMEDOUT;
		}
	}

	DEBUG(2, "\n");
	DEBUG(1, "ADwin debug: read %d bytes from Transputer\n", (__u8*)ptr - (__u8*)usr_buf);

        return (__u8 *)ptr - (__u8 *)usr_buf;
}

/*****************************************************************************/
/* write to link (transputer)
 */
static int to_link(wait_queue_head_t *q, int base_addr,
					const char *usr_buf, size_t nbytes) {
        int i;
	__u8 daten, *ptr = (__u8 *) usr_buf;
	struct timer_list timer;
	volatile long timeout_flag = 0;

	if (! access_ok(VERIFY_READ, usr_buf, nbytes)) {
		DEBUG(0, "ADwin debug: Memory access error\n");
		return -EFAULT;
	}

	DEBUG(2, "ADwin debug: ");

	/* same procesdure as before */
        while (nbytes > 0) {
		__get_user(daten, ptr++);
		DEBUG(2, "%02X ", daten);

		for (i = 0; (i < 100) && WRITE_STATUS(base_addr); i++);

		if (i < 100) {
			outb(daten, base_addr + 1);
			nbytes--;
			continue;
		}

		init_timer(&timer);
		timer.expires  = jiffies + 25;
		timer.data     = (unsigned long) &timeout_flag;
		timer.function = adwin_timeout;

		add_timer(&timer);
		while(! timeout_flag) {
			if (WRITE_STATUS(base_addr) != -EACCES) {
				outb(daten, base_addr + 1);
				nbytes--;
				break;
			} else
				schedule();
		}
		del_timer_sync(&timer);

		if (timeout_flag) {
			DEBUG(1, "\nADwin debug: timed out\n");
			return -ETIMEDOUT;
		}
	}

	DEBUG(1, "Adwin debug: wrote %d bytes to Transputer\n",
						(__u8 *)ptr - (__u8 *)usr_buf);

	return ((__u8 *)ptr - (__u8 *)usr_buf);
}

/*****************************************************************************/
/* read from link (DSP)
 */
static int from_link_dsp(wait_queue_head_t *q, int base_addr,
					const char *usr_buf, size_t cnt) {
        int i, j;
 	__u32 daten, *lptr = (__u32 *)usr_buf;
	__u8 *byte = (__u8 *) &daten;
	struct timer_list timer;
	volatile long timeout_flag = 0;				

	if (! access_ok(VERIFY_WRITE, usr_buf, cnt)) {
		DEBUG(1, "ADwin debug: Memory access error\n");
		return -EFAULT;
	}

	DEBUG(2, "ADwin debug: ");

        while(cnt > 0) {
		for (j = 0; j < 4; j++) {
			for (i = 0; (i < 100) && READ_STATUS(base_addr); i++);

			if (i < 100) {
				byte[j] = inb(base_addr);
				DEBUG(2, "%02X ", byte[j]);
				cnt--;
				continue;
			}

			init_timer(&timer);
			timer.expires  = jiffies + 25;
			timer.data     = (unsigned long) &timeout_flag;
			timer.function = adwin_timeout;

			add_timer(&timer);
			while(! timeout_flag) {
				if (READ_STATUS(base_addr) != -EACCES) {
					byte[j] = inb(base_addr);
					DEBUG(2, "%02X ", byte[j]);
					cnt--;
					break;
				} else
					schedule();
			}
			del_timer_sync(&timer);

			if (timeout_flag) {
				DEBUG(2, "\n");
				DEBUG(1, "ADwin debug: timed out\n");
				return -ETIMEDOUT;
			}
		}

		/* SHARC big endian to CPU (Intel, ARM) byte order */
		__put_user(be32_to_cpu(daten), lptr++);
	}
	DEBUG(2, "\n");
	DEBUG(1, "ADwin debug: read %d bytes from DSP\n", 
					(__u8 *)lptr - (__u8 *)usr_buf);

        return ((__u8 *)lptr - (__u8 *)usr_buf);
}

/*****************************************************************************/
/* write to link (DSP)
 */
static int to_link_dsp(wait_queue_head_t *q, int base_addr,
					const char *usr_buf, size_t nbytes) {

	int j;
	__u32 daten, *lptr = (__u32 *)usr_buf;
	__u8 *byte = (__u8 *)&daten;

	if (! access_ok(VERIFY_READ, usr_buf, nbytes)) {
		DEBUG(1, "ADwin debug: Memory access error\n");
		return -EFAULT;
	}

	DEBUG(2, "ADwin debug: ");

	/* DSP-Communication is word-wise (4byte) */
        while (nbytes > 0) {
		__get_user(daten, lptr++);
		/* CPU to SHARC big endian byte order */
		daten = cpu_to_be32(daten);

		for (j = 0; j < 4; j++) {
			if (__to_link_dsp(byte + j, base_addr))
				return -ETIMEDOUT;

			nbytes--;
		}
	}

	DEBUG(2, "\n");
	DEBUG(1, "\nADwin debug: wrote %d bytes to DSP\n", (__u8 *)lptr - (__u8 *)usr_buf);

	return ((__u8 *)lptr - (__u8 *)usr_buf);
}


/*****************************************************************************/
int adwin_read(struct file *filp, char *usr_buf, size_t count, loff_t *ppos) {
	adwin_kdev_t *kdev = (adwin_kdev_t *) filp->private_data;

	switch(kdev->kdev_processor) {
		case DP_DSP:
			return from_link_dsp(&kdev->kdev_queue,
					     kdev->kdev_link_base_addr,
					     usr_buf,
					     count);
		case DP_T2:
		case DP_T4:
		case DP_T8:
			return from_link(&kdev->kdev_queue,
					 kdev->kdev_link_base_addr,
					 usr_buf,
					 count);
		default:
			printk(KERN_WARNING MODULE_NAME ": protocol/processor unknown\n");
			return -EINVAL;
	}
}

/*****************************************************************************/
int adwin_write(struct file *filp, const char *usr_buf, size_t count, loff_t *ppos) {
	adwin_kdev_t *kdev = (adwin_kdev_t *) filp->private_data;

	switch(kdev->kdev_processor) {
		case DP_DSP:
			return to_link_dsp(&kdev->kdev_queue,
					     kdev->kdev_link_base_addr,
					     usr_buf,
					     count);
		case DP_T2:
		case DP_T4:
		case DP_T8:
			return to_link(&kdev->kdev_queue,
					 kdev->kdev_link_base_addr,
					 usr_buf,
					 count);
		default:
			printk(KERN_WARNING MODULE_NAME ": protocol/processor unknown\n");
			return -EINVAL;
	}
}
