#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#ifdef ADWIN_MODVERSIONS
#include <linux/modversions.h>
#endif

#include <linux/kernel.h>		/* we live in kernel */
#include <linux/module.h>		/* we are a module */
#include <linux/init.h>
#include <linux/spinlock.h>		/* SMP */
#include <linux/stddef.h>		/* offsetof() */
#include <linux/list.h>			/* linked list */

#include <asm/uaccess.h>		/* access User-Space */
#include <asm/atomic.h>			/* atomic operations */

#include <linux/adwin.h>

MODULE_AUTHOR(PACKAGE_BUGREPORT);
MODULE_DESCRIPTION("ADwin Link Driver");

#ifdef MODULE_LICENSE
  #ifdef WITH_PCMCIA
    MODULE_LICENSE("Dual MPL/GPL");
  #else
    MODULE_LICENSE("GPL");
  #endif
#endif

#ifdef ADWIN_DEBUG
  static int debug = ADWIN_DEBUG;
  MODULE_PARM(debug, "i");
  #define DEBUG(n, args...) if (debug > (n)) printk(args)
#else
  #define DEBUG(n, args...)
#endif

/* definition of DEBUG is needed */
#ifdef WITH_LINK
  #include "adwin_link.h"
#endif

/* locally defined functions */
static int adwin_open(struct inode *inode, struct file *filp);
static int adwin_release(struct inode *inode, struct file *filp);
static int adwin_ioctl(struct inode *inode, struct file *filp,
                                unsigned int cmd, unsigned long arg);

/* defined in adwin_proc.c */
extern int proc_devices_read(char *page, char **start, off_t off,
			  int count, int *eof, void *data);
extern int release_adwin_resource(int base_addr);
extern struct file_operations proc_config_fops;

int proc_dev_create (struct proc_dir_entry *parent, adwin_kdev_t *kdev);

/* defined in adwin_cs.c */
int init_adwin_cs(void);
int exit_adwin_cs(void);


/* These are pointers to the directories in /proc
 * We keep them in mind, so thay are easier to access
 */
struct proc_dir_entry *adwin_root = NULL;
struct proc_dir_entry *dev_root   = NULL;

/* File opertaion structures
 * Listed operations are supported
 */
static struct file_operations adwin_fops = {
	open:		adwin_open,
	release:	adwin_release,
	ioctl:		adwin_ioctl,
#ifdef WITH_LINK
	read:		adwin_read,
	write:		adwin_write,
#endif
};


/*****************************************************************************/
static struct {
	int out, in, offset, len;
} iomemmap[] = {
	{ ADIOGKDEVINFO, ADIOSKDEVINFO,
		offsetof(adwin_kdev_t, dev), sizeof(adwin_dev_t) },
	{ ADIOGKDEVDESC, ADIOSKDEVDESC,
		offsetof(adwin_kdev_t, kdev_desc), 32 },
	{ ADIOGKDEVUID, ADIOSKDEVUID,
		offsetof(adwin_kdev_t, kdev_uid), sizeof(uid_t) },
	{ ADIOGKDEVGID, ADIOSKDEVGID,
		offsetof(adwin_kdev_t, kdev_gid), sizeof(gid_t) },
	{ ADIOGKDEVMODE, ADIOSKDEVMODE,
		offsetof(adwin_kdev_t, kdev_mode), sizeof(mode_t) },
	{ ADIOGKDEVTYPE, ADIOSKDEVTYPE,
		offsetof(adwin_kdev_t, kdev_type), sizeof(int) },
	{ ADIOGKDEVBASE, ADIOSKDEVBASE,
		offsetof(adwin_kdev_t, kdev_link_base_addr), sizeof(int) },
	{ ADIOGKDEVSERIAL, ADIOSKDEVSERIAL,
		offsetof(adwin_kdev_t, kdev_usb_serial), sizeof(int) },
	{ ADIOGKDEVSERVER, ADIOSKDEVSERVER,
		offsetof(adwin_kdev_t, kdev_net_server), sizeof(struct sockaddr_in) },
	{ ADIOGKDEVTIMEO, ADIOSKDEVTIMEO,
		offsetof(adwin_kdev_t, kdev_net_timeout), sizeof(int) },
	{ ADIOGKDEVCOUNT, ADIOSKDEVCOUNT,
		offsetof(adwin_kdev_t, kdev_net_package_count), sizeof(int) },
	{ ADIOGKDEVHSTLNK, ADIOSKDEVHSTLNK,
		offsetof(adwin_kdev_t, kdev_net_host_link), sizeof(int) },
	{ ADIOGKDEVPW, ADIOSKDEVPW,
		offsetof(adwin_kdev_t, kdev_net_host_password), 12 },
	{ 0, 0, 0, 0 },
};

/*****************************************************************************/
static int adwin_ioctl(struct inode *inode, struct file *filp,
				unsigned int cmd, unsigned long arg) {
	adwin_kdev_t *kdev = (adwin_kdev_t *) filp->private_data;
	DEBUG(1, "ADwin debug: adwin_ioctl() - ");

	switch(cmd) {
#ifdef WITH_LINK
		case ADIORESETLINK:
			DEBUG(1, "ADIORESETLINK\n");
			return reset_link(&kdev->kdev_queue, kdev->kdev_link_base_addr);

		case ADIOGREADSTATUS:
			DEBUG(1, "ADIOGREADSTATUS: %d\n",
						READ_STATUS(kdev->kdev_link_base_addr));
			return READ_STATUS(kdev->kdev_link_base_addr);

		case ADIOGWRITESTATUS:
			DEBUG(1, "ADIOGWRITESTATUS: %d\n",
						WRITE_STATUS(kdev->kdev_link_base_addr));
			return WRITE_STATUS(kdev->kdev_link_base_addr);
#endif /* WITH_LINK */
		case ADIOSKDEVPROC:
			DEBUG(1, "ADIOSKDEVPROC: %d\n", (int)arg);
			kdev->kdev_processor = arg;
			return 0;
#ifdef WITH_LINK
		case ADIOSOPCODE:
			{
			/* write 6 opcode-bytes linear to DSP */
			int i;
			char opcode[6];
			copy_from_user(opcode, (long *)arg, 6);
			DEBUG(1, "ADIOSOPCODE\n");
			DEBUG(2, "ADwin debug: ");
			for (i = 0; i < 6; i++) {
				if (__to_link_dsp(opcode + i, kdev->kdev_link_base_addr) < 0)
					return -ETIMEDOUT;
			}
			return 0;
			}
#endif /* WITH_LINK */
	}

	if (_IOC_DIR(cmd) & _IOC_READ) {
		int i;
		for(i = 0; iomemmap[i].out; i++)
			if (cmd == iomemmap[i].out) {
/*
printk("ioctl(): iomemmap.offset = %d\n", iomemmap[i].offset);
printk("ioctl(): iomemmap.len = %d\n", iomemmap[i].len);
printk("ioctl(): (char *)arg = %X\n", (char *)arg);
printk("ioctl(): arg = %X\n", arg);
printk("ioctl(): kdev = %X\n", kdev);
printk("ioctl(): ((char *)kdev) + iomemmap[i].offset = %X\n", ((char *)kdev) + iomemmap[i].offset);
*/
				return copy_to_user((char*)arg,
							((char *)kdev) + iomemmap[i].offset,
							iomemmap[i].len);
			}

	} else if (_IOC_DIR(cmd) & _IOC_WRITE){
		int i;

		if (! capable(CAP_SYS_ADMIN))
			return -EPERM;

		for(i = 0; iomemmap[i].in; i++)
			if (cmd == iomemmap[i].in) {
				return copy_from_user((char *)arg + iomemmap[i].offset,
							((long *)arg),
							iomemmap[i].len);

			}
	}

	return -ENOTTY;
}

/*****************************************************************************/
static int adwin_release(struct inode *inode, struct file *filp) {
	adwin_kdev_t *kdev;
	spinlock_t adwin_lock;
	spin_lock_init(&adwin_lock);
	kdev = (adwin_kdev_t *) filp->private_data;

	DEBUG(1, "ADwin debug: adwin_release(0x%X)\n", kdev->kdev_link_base_addr);

	spin_lock(&adwin_lock);
	atomic_dec(&kdev->kdev_in_use);
	atomic_inc(&kdev->kdev_use_count);
	MOD_DEC_USE_COUNT;
	spin_unlock(&adwin_lock);

	/* wake up any process which might be around and watch the race ;)
	 */
	wake_up_interruptible(&kdev->kdev_queue);
	return 0;
}

/*****************************************************************************/
/* A few words may be necessary:
 * Only ONE process at a time is allowed to access a device, since we
 * have synchronous link communication.
 * Therefore we have to suspend every process trying to access a device if
 * it's already in use (kdev->kdev_in_use != 0).
 *
 */
static int adwin_open(struct inode *inode, struct file *filp) {
	adwin_kdev_t *kdev;
	spinlock_t adwin_lock;
	struct proc_dir_entry * ent;

	ent = (struct proc_dir_entry *)inode->u.generic_ip;
	kdev = (adwin_kdev_t *)ent->data;

	/* lock the critical section */
	spin_lock_init(&adwin_lock);

	spin_lock(&adwin_lock);
	MOD_INC_USE_COUNT;

	while(atomic_read(&kdev->kdev_in_use)) {
		spin_unlock(&adwin_lock);

		/* this operation would SURELY block */
		if (filp->f_flags & O_NONBLOCK) {
			MOD_DEC_USE_COUNT;
			return -EAGAIN; /* EWOULDBLOCK */
		}

		/* device in use, let's take a nap until device is
		 * available again and someone wakes us up
		 * (wake_up_interruptible() in adwin_release())
		 */
		interruptible_sleep_on(&kdev->kdev_queue);

		/* I'm not sure about this myself, but the book
		 * I got it from was almost always right ;)
		 */
		if(signal_pending(current)) {
			MOD_DEC_USE_COUNT;
			return -ERESTARTSYS;
		}

		spin_lock(&adwin_lock);
	}

	/* grab the "in use" flag. */
	atomic_inc(&kdev->kdev_in_use);

	/* every function, read, write, ioctl and release will find
	 * the data in this pointer.
	 */
	filp->private_data = kdev;
	spin_unlock(&adwin_lock);

	return 0;
}

/*****************************************************************************/
/* register the character device and
 * init the /proc/adwin directory,
 */
static int __init adwin_init(void) {
	int e;
	struct proc_dir_entry *config, *devices;

	if ((e = register_chrdev(ADWIN_MAJOR, MODULE_NAME, &adwin_fops)) < 0) {
		printk(KERN_ERR "ADwin init: major %d not available\n",
								ADWIN_MAJOR);
		return e;
	}

	EXPORT_NO_SYMBOLS;

	printk(KERN_NOTICE "ADwin init: %s %s %s\n", 
	       VERSION, VERSION_DATE, PACKAGE_BUGREPORT);

	DEBUG(1, "ADwin debug: creating /proc/adwin ... ");
        adwin_root           = proc_mkdir(MODULE_NAME, &proc_root);
	adwin_root->owner    = THIS_MODULE;
	dev_root             = proc_mkdir("dev", adwin_root);
	dev_root->owner      = THIS_MODULE;

	devices = create_proc_entry("devices", S_IFREG | 0444, adwin_root);
	devices->read_proc   = proc_devices_read;
	devices->owner       = THIS_MODULE;

	config               = create_proc_entry("config", S_IFREG | 0640,
							adwin_root);
	config->proc_fops    = &proc_config_fops;
	config->owner        = THIS_MODULE;
	DEBUG(1, "done\n");

#ifdef WITH_PCMCIA
	init_adwin_cs();
#endif
	
	return 0;
}

/*****************************************************************************/
/* unregister and release
 */
static void __exit adwin_cleanup(void) {
	struct proc_dir_entry	*p;

	unregister_chrdev(ADWIN_MAJOR, MODULE_NAME);

	DEBUG(1, "ADwin debug: /proc/adwin/dev ... ");
	for(p = dev_root->subdir; p; p = p->next) {
		adwin_kdev_t *kdev = (adwin_kdev_t *)p->data;

		if (kdev->kdev_type == DC_LINK) {
			release_adwin_resource(kdev->kdev_link_base_addr);
			release_adwin_resource(kdev->kdev_link_base_addr + 0x10);
		}

		kfree(p->data);
	}

	remove_proc_entry("dev", adwin_root);
	DEBUG(1, "destroyed\n");

	remove_proc_entry("devices", adwin_root);
	remove_proc_entry("adwin", &proc_root);

#ifdef	WITH_PCMCIA
	exit_adwin_cs();
#endif
}

module_init(adwin_init);
module_exit(adwin_cleanup);
